#!/usr/bin/env python3

"""
Intra2net 2017

===============================================================================
              crypto -- Encryption Layer for the Intra2net Backup
===============================================================================

Crypto stack:

    - AES-GCM for the symmetric encryption;
    - Scrypt as KDF.

References:

    - NIST Recommendation for Block Cipher Modes of Operation: Galois/Counter
      Mode (GCM) and GMAC
      http://nvlpubs.nist.gov/nistpubs/Legacy/SP/nistspecialpublication800-38d.pdf

    - AES-GCM v1:
      https://cryptome.org/2014/01/aes-gcm-v1.pdf

    - Authentication weaknesses in GCM
      http://csrc.nist.gov/groups/ST/toolkit/BCM/documents/comments/CWC-GCM/Ferguson2.pdf

Trouble with python-cryptography packages: authentication tags can only be
passed in advance: https://github.com/pyca/cryptography/pull/3421

Errors
-------------------------------------------------------------------------------

Errors fall into roughly three categories:

    - Cryptographical errors or data validity.

        - ``InvalidGCMTag`` (decryption failed on account of an invalid GCM
          tag),
        - ``InvalidIVFixedPart`` (IV fixed part of object not found in list),
        - ``DuplicateIV`` (the IV of an encrypted object already occurred),
        - ``DecryptionError`` (used in CLI decryption).

    - Incorrect usage of the library.

        - ``InvalidParameter`` (non-conforming user supplied parameter),
        - ``InvalidHeader`` (data passed for reading not parsable into header),
        - ``FormatError`` (cannot handle header or parameter version),
        - ``RuntimeError``.

    - Bad internal state. If one of these is encountered it means that a state
      was reached that shouldn’t occur during normal processing.

        - ``InternalError``,
        - ``Unreachable``.

Also, ``EndOfFile`` is used as a sentinel to communicate that a stream supplied
for reading is exhausted.

Initialization Vectors
-------------------------------------------------------------------------------

Initialization vectors are checked reuse during the lifetime of a decryptor.

"""

import binascii
import bisect
import ctypes
import io
from functools import reduce, partial
import os
import struct
import sys
import time
import types
try:
    import enum34
except ImportError as exn:
    pass

if __name__ == "__main__": ## Work around the import mechanism’s lest Python’s
    pwd = os.getcwd()      ## preference for local imports causes a cyclical
    ## import (crypto → pylibscrypt → […] → ./tarfile → crypto).
    sys.path = [ p for p in sys.path if p.find ("deltatar") < 0 ]

import pylibscrypt
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
import cryptography


__all__ = [ "hdr_make", "hdr_read", "hdr_fmt", "hdr_fmt_pretty"
          , "scrypt_hashfile"
          , "PDTCRYPT_HDR_SIZE", "AES_GCM_IV_CNT_DATA"
          , "AES_GCM_IV_CNT_INFOFILE", "AES_GCM_IV_CNT_INDEX"
          ]


###############################################################################
## exceptions
###############################################################################

class EndOfFile (Exception):
    """Reached EOF."""
    remainder = 0
    msg       = 0
    def __init__ (self, n=None, msg=None):
        if n is not None:
            self.remainder = n
        self.msg = msg


class InvalidParameter (Exception):
    """Inputs not valid for PDT encryption."""
    pass


class InvalidHeader (Exception):
    """Header not valid."""
    pass


class InvalidGCMTag (Exception):
    """
    The GCM tag calculated during decryption differs from that in the object
    header.
    """
    pass


class InvalidIVFixedPart (Exception):
    """IV fixed part not in supplied list."""
    pass


class DuplicateIV (Exception):
    """IV reused."""
    pass


class NonConsecutiveIV (Exception):
    """IVs not numbered consecutively (not necessarily an error)."""
    pass


class FormatError (Exception):
    """Unusable parameters in header."""
    pass


class DecryptionError (Exception):
    """Error during decryption."""
    pass


class Unreachable (Exception):
    """Makeshift __builtin_unreachable()."""
    pass


class InternalError (Exception):
    """Errors not ascribable to bad user inputs or cryptography."""
    pass


###############################################################################
## crypto layer version
###############################################################################

ENCRYPTION_PARAMETERS = \
    { 0: \
        { "kdf": ("dummy", 16)
        , "enc": "passthrough" }
    , 1: \
        { "kdf": ( "scrypt"
                 , { "dkLen"    : 16
                   , "N"        : 1 << 16
                   , "r"        : 8
                   , "p"        : 1
                   , "NaCl_LEN" : 16 })
        , "enc": "aes-gcm" } }

###############################################################################
## constants
###############################################################################

PDTCRYPT_HDR_MAGIC = b"PDTCRYPT"

PDTCRYPT_HDR_SIZE_MAGIC          = 8     #  8
PDTCRYPT_HDR_SIZE_VERSION        = 2     # 10
PDTCRYPT_HDR_SIZE_PARAMVERSION   = 2     # 12
PDTCRYPT_HDR_SIZE_NACL           = 16    # 28
PDTCRYPT_HDR_SIZE_IV             = 12    # 40
PDTCRYPT_HDR_SIZE_CTSIZE         = 8     # 48
PDTCRYPT_HDR_SIZE_TAG            = 16    # 64 GCM auth tag

PDTCRYPT_HDR_SIZE = PDTCRYPT_HDR_SIZE_MAGIC        + PDTCRYPT_HDR_SIZE_VERSION \
                  + PDTCRYPT_HDR_SIZE_PARAMVERSION + PDTCRYPT_HDR_SIZE_NACL    \
                  + PDTCRYPT_HDR_SIZE_IV           + PDTCRYPT_HDR_SIZE_CTSIZE  \
                  + PDTCRYPT_HDR_SIZE_TAG # = 64

# precalculate offsets since Python can’t do constant folding over names
HDR_OFF_VERSION      = PDTCRYPT_HDR_SIZE_MAGIC
HDR_OFF_PARAMVERSION = HDR_OFF_VERSION      + PDTCRYPT_HDR_SIZE_VERSION
HDR_OFF_NACL         = HDR_OFF_PARAMVERSION + PDTCRYPT_HDR_SIZE_PARAMVERSION
HDR_OFF_IV           = HDR_OFF_NACL         + PDTCRYPT_HDR_SIZE_NACL
HDR_OFF_CTSIZE       = HDR_OFF_IV           + PDTCRYPT_HDR_SIZE_IV
HDR_OFF_TAG          = HDR_OFF_CTSIZE       + PDTCRYPT_HDR_SIZE_CTSIZE

FMT_UINT16_LE = "<H"
FMT_UINT64_LE = "<Q"
FMT_I2N_IV    = "<8sL"   # 8 random bytes ‖ 32 bit counter
FMT_I2N_HDR   = ("<"     # host byte order
                 "8s"    # magic
                 "H"     # version
                 "H"     # paramversion
                 "16s"   # sodium chloride
                 "12s"   # iv
                 "Q"     # size
                 "16s")  # GCM tag

# aes+gcm
AES_GCM_MAX_SIZE              = (1 << 36) - (1 << 5) # 2^39 - 2^8 b ≅ 64 GB
PDTCRYPT_MAX_OBJ_SIZE_DEFAULT = 63 * (1 << 30)       #                63 GB
PDTCRYPT_MAX_OBJ_SIZE         = PDTCRYPT_MAX_OBJ_SIZE_DEFAULT

# index and info files are written on-the fly while encrypting so their
# counters must be available inadvance
AES_GCM_IV_CNT_INFOFILE     = 1 # constant
AES_GCM_IV_CNT_INDEX        = AES_GCM_IV_CNT_INFOFILE + 1
AES_GCM_IV_CNT_DATA         = AES_GCM_IV_CNT_INDEX    + 1 # also for multivolume
AES_GCM_IV_CNT_MAX_DEFAULT  = 0xffFFffFF
AES_GCM_IV_CNT_MAX          = AES_GCM_IV_CNT_MAX_DEFAULT


###############################################################################
## header, trailer
###############################################################################
#
# Interface:
#
#    struct hdrinfo
#      { version      : u16
#      , paramversion : u16
#      , nacl         : [u8; 16]
#      , iv           : [u8; 12]
#      , ctsize       : usize }
#
#    tag : [u8; 16]
#
#    fn hdr_read (f : handle) -> hdrinfo;
#    fn hdr_make (f : handle, h : hdrinfo) -> IOResult<usize>;
#    fn hdr_fmt (h : hdrinfo) -> String;
#

def hdr_read (data):

    try:
        mag, version, paramversion, nacl, iv, ctsize, tag = \
            struct.unpack (FMT_I2N_HDR, data)
    except Exception as exn:
        raise InvalidHeader ("error unpacking header from [%r]: %s"
                             % (binascii.hexlify (data), str (exn)))

    if mag != PDTCRYPT_HDR_MAGIC:
        raise InvalidHeader ("bad magic in header: expected [%s], got [%s]"
                             % (PDTCRYPT_HDR_MAGIC, mag))

    return \
        {      "version" : version
        , "paramversion" : paramversion
        ,         "nacl" : nacl
        ,           "iv" : iv
        ,       "ctsize" : ctsize
        ,          "tag" : tag
        }


def hdr_read_stream (instr):
    data = instr.read(PDTCRYPT_HDR_SIZE)
    ldata = len (data)
    if ldata == 0:
        raise EndOfFile
    elif ldata != PDTCRYPT_HDR_SIZE:
        raise InvalidHeader ("hdr_read_stream: expected %d B, received %d B"
                             % (PDTCRYPT_HDR_SIZE, ldata))
    return hdr_read (data)


def hdr_from_params (version, paramversion, nacl, iv, ctsize, tag):
    buf  = bytearray (PDTCRYPT_HDR_SIZE)
    bufv = memoryview (buf)

    try:
        struct.pack_into (FMT_I2N_HDR, bufv, 0,
                          PDTCRYPT_HDR_MAGIC,
                          version, paramversion, nacl, iv, ctsize, tag)
    except Exception as exn:
        return False, "error writing header: %s" % str (exn)

    return True, bytes (buf)


def hdr_make_dummy (s):
    """
    Create a header sized block of bytes initialized to a value derived from a
    string. Used to verify we’ve jumped back correctly to the actual position
    of the object header.
    """
    c = reduce (lambda a, c: a + ord(c), s, 0) % 0xFF
    return bytes (bytearray (struct.pack ("B", c)) * PDTCRYPT_HDR_SIZE)


def hdr_make (hdr):
    return hdr_from_params (version=hdr.get("version"),
                            paramversion=hdr.get("paramversion"),
                            nacl=hdr.get("nacl"), iv=hdr.get("iv"),
                            ctsize=hdr.get("ctsize"), tag=hdr.get("tag"))


HDR_FMT = "I2n_header { version: %d, paramversion: %d, nacl: %s[%d]," \
                      " iv: %s[%d], ctsize: %d, tag: %s[%d] }"

def hdr_fmt (h):
    return HDR_FMT % (h["version"], h["paramversion"],
                      binascii.hexlify (h["nacl"]), len(h["nacl"]),
                      binascii.hexlify (h["iv"]), len(h["iv"]),
                      h["ctsize"],
                      binascii.hexlify (h["tag"]), len(h["tag"]))


def hex_spaced_of_bytes (b):
    return " ".join ([ "%.2x%.2x" % (c1, c2)
                       for c1, c2 in zip (b[0::2], b[1::2]) ]) \
         + (len (b) | 1 == len (b) and " %.2x" % b[-1] or "") # odd lengths


def hdr_iv_counter (h):
    """Extract the variable part of the IV of the given header."""
    _fixed, cnt = struct.unpack (FMT_I2N_IV, h ["iv"])
    return cnt


def hdr_iv_fixed (h):
    """Extract the fixed part of the IV of the given header."""
    fixed, _cnt = struct.unpack (FMT_I2N_IV, h ["iv"])
    return fixed


hdr_dump = hex_spaced_of_bytes


HDR_FMT_PRETTY = \
"""version         = %-4d                 : %s
paramversion    = %-4d                 : %s
nacl                                   : %s
iv                                     : %s
ctsize          = %-20d : %s
tag                                    : %s
"""

def hdr_fmt_pretty (h):
    return HDR_FMT_PRETTY \
                % (h["version"],
                   hex_spaced_of_bytes (struct.pack (FMT_UINT16_LE, h["version"])),
                   h["paramversion"],
                   hex_spaced_of_bytes (struct.pack (FMT_UINT16_LE, h["paramversion"])),
                   hex_spaced_of_bytes (h["nacl"]),
                   hex_spaced_of_bytes (h["iv"]),
                   h["ctsize"],
                   hex_spaced_of_bytes (struct.pack (FMT_UINT64_LE, h["ctsize"])),
                   hex_spaced_of_bytes (h["tag"]))

IV_FMT = "((f %s) (c %d))"

def iv_fmt (iv):
    fixed, cnt = struct.unpack (FMT_I2N_IV, iv)
    return IV_FMT % (binascii.hexlify (fixed), cnt)


###############################################################################
## passthrough / null encryption
###############################################################################

class PassthroughCipher (object):

    tag = struct.pack ("<QQ", 0, 0)

    def __init__          (self)    : pass

    def update            (self, b) : return b

    def finalize          (self)    : return b""

    def finalize_with_tag (self, _) : return b""

###############################################################################
## convenience wrapper
###############################################################################


def kdf_dummy (klen, password, _nacl):
    q, r = divmod (klen, len (password))
    if isinstance (password, bytes) is False:
        password = password.encode ()
    return password * q + password [:r], b""


SCRYPT_KEY_MEMO = { } # static because needed for both the info file and the archive


def kdf_scrypt (params, password, nacl):
    N = params["N"]
    r = params["r"]
    p = params["p"]
    dkLen = params["dkLen"]

    if nacl is None:
        nacl = os.urandom (params["NaCl_LEN"])

    key_parms = (password, nacl, N, r, p, dkLen)
    global SCRYPT_KEY_MEMO
    if key_parms not in SCRYPT_KEY_MEMO:
        SCRYPT_KEY_MEMO [key_parms] = \
            pylibscrypt.scrypt (password, nacl, N, r, p, dkLen)
    return SCRYPT_KEY_MEMO [key_parms], nacl


def kdf_by_version (paramversion=None, defs=None):
    if paramversion is not None:
        defs = ENCRYPTION_PARAMETERS.get(paramversion, None)
    if defs is None:
        raise ValueError ("no encryption parameters for version %r"
                % paramversion)
    (kdf, params) = defs["kdf"]
    fn = None
    if kdf == "scrypt" : fn = kdf_scrypt
    if kdf == "dummy"  : fn = kdf_dummy
    if fn is None:
        raise ValueError ("key derivation method %r unknown" % kdf)
    return partial (fn, params)


###############################################################################
## SCRYPT hashing
###############################################################################

def scrypt_hashsource (pw, ins):
    """
    Calculate the SCRYPT hash from the password and the information contained
    in the first header found in ``ins``.

    This does not validate whether the first object is encrypted correctly.
    """
    hdr = None
    try:
        hdr = hdr_read_stream (ins)
    except EndOfFile as exn:
        noise ("PDT: malformed input: end of file reading first object header")
        noise ("PDT:")
        return 1

    nacl = hdr ["nacl"]
    pver = hdr ["paramversion"]
    if PDTCRYPT_VERBOSE is True:
        noise ("PDT: salt of first object          : %s" % binascii.hexlify (nacl))
        noise ("PDT: parameter version of archive  : %d" % pver)

    try:
        defs = ENCRYPTION_PARAMETERS.get(pver, None)
        kdfname, params = defs ["kdf"]
        if kdfname != "scrypt":
            noise ("PDT: input is not an SCRYPT archive")
            noise ("")
            return 1
        kdf = kdf_by_version (None, defs)
    except ValueError as exn:
        noise ("PDT: object has unknown parameter version %d" % pver)

    hsh, _void = kdf (pw, nacl)

    return hsh, nacl, params


def scrypt_hashfile (pw, fname):
    with deptdcrypt_mk_stream (PDTCRYPT_SOURCE, fname  or "-") as ins:
        hsh, _void, _void = scrypt_hashsource (pw, ins)
        return hsh


###############################################################################
## AES-GCM context
###############################################################################

class Crypto (object):
    """
    Encryption context to remain alive throughout an entire tarfile pass.
    """
    enc  = None
    nacl = None
    key  = None
    cnt  = None # file counter (uint32_t != 0)
    iv   = None # current IV
    fixed        = None  # accu for 64 bit fixed parts of IV
    used_ivs     = None  # tracks IVs
    strict_ivs   = False # if True, panic on duplicate object IV
    password     = None
    paramversion = None
    stats = { "in"  : 0
            , "out" : 0
            , "obj" : 0 }

    ctsize  = -1
    ptsize  = -1
    info_counter_used  = False
    index_counter_used = False

    def __init__ (self, *al, **akv):
        self.used_ivs = set ()
        self.set_parameters (*al, **akv)


    def next_pfx (self):
        pass


    def set_object_counter (self, cnt=None):
        if cnt is None:
            self.cnt = AES_GCM_IV_CNT_DATA
            return
        if cnt == 0 or cnt > AES_GCM_IV_CNT_MAX + 1:
            raise InvalidParameter ("invalid counter value %d requested: "
                                    "acceptable values are from 1 to %d"
                                    % (cnt, AES_GCM_IV_CNT_MAX))
        if cnt == AES_GCM_IV_CNT_INFOFILE:
            if self.info_counter_used is True:
                raise InvalidParameter ("attempted to reuse info file counter "
                                        "%d: must be unique" % cnt)
            self.info_counter_used = True
        elif cnt == AES_GCM_IV_CNT_INDEX:
            if self.index_counter_used is True:
                raise InvalidParameter ("attempted to reuse index file counter "
                                        "%d: must be unique" % cnt)
            self.index_counter_used = True
        if cnt <= AES_GCM_IV_CNT_MAX:
            self.cnt = cnt
            return
        # cnt == AES_GCM_IV_CNT_MAX + 1 → wrap
        self.cnt = AES_GCM_IV_CNT_DATA
        self.next_pfx ()


    def set_parameters (self, password=None, key=None, paramversion=None,
                        nacl=None, counter=None, nextpfx=None,
                        strict_ivs=False):
        if nextpfx is not None:
            self.next_pfx = nextpfx
            self.next_pfx ()
        self.set_object_counter (counter)

        self.strict_ivs = strict_ivs

        if key is not None:
            self.key, self.nacl = key, nacl
            return

        if isinstance (password, bytes) is False: password = str.encode (password)
        self.password = password
        if paramversion is None and nacl is None:
            # postpone until first header is available
            return
        kdf = kdf_by_version (paramversion)
        if kdf is not None:
            self.key, self.nacl = kdf (password, nacl)
        self.paramversion = paramversion


    def process (self, buf):
        if self.enc is None:
            raise RuntimeError ("process: context not initialized")
        self.stats ["in"] += len (buf)
        out = self.enc.update (buf)
        self.stats ["out"] += len (out)
        return out


    def next (self, password, paramversion, nacl, iv):
        self.ctsize = 0
        self.ptsize = 0
        self.stats ["obj"] += 1

        self.check_duplicate_iv (iv)

        if (   self.paramversion != paramversion
            or self.password     != password
            or self.nacl         != nacl):
            self.set_parameters (password=password, paramversion=paramversion,
                                 nacl=nacl, strict_ivs=self.strict_ivs)


    def check_duplicate_iv (self, iv):
        if self.strict_ivs is True and iv in self.used_ivs:
            raise DuplicateIV ("iv %s was reused" % iv_fmt (iv))
        # vi has not been used before; add to collection
        self.used_ivs.add (iv)


    def counters (self):
        return self.stats ["obj"], self.stats ["in"], self.stats ["out"]


class Encrypt (Crypto):

    lastinfo     = None
    version      = None
    paramenc     = None

    def __init__ (self, version, paramversion, password=None, key=None, nacl=None,
                  counter=AES_GCM_IV_CNT_DATA, strict_ivs=True):
        if         password is     None and key is     None \
                or password is not None and key is not None :
            raise InvalidParameter ("__init__: need either key or password")

        if key is not None:
            if isinstance (key, bytes) is False:
                raise InvalidParameter ("__init__: key must be provided as "
                                        "bytes, not %s" % type (key))
            if nacl is None:
                raise InvalidParameter ("__init__: salt must be provided along "
                                        "with encryption key")
        else: # password, no key
            if isinstance (password, str) is False:
                raise InvalidParameter ("__init__: password must be a string, not %s"
                                        % type (password))
            if len (password) == 0:
                raise InvalidParameter ("__init__: supplied empty password but not "
                                        "permitted for PDT encrypted files")
        # version
        if isinstance (version, int) is False:
            raise InvalidParameter ("__init__: version number must be an "
                                    "integer, not %s" % type (version))
        if version < 0:
            raise InvalidParameter ("__init__: version number must be a "
                                    "nonnegative integer, not %d" % version)
        # paramversion
        if isinstance (paramversion, int) is False:
            raise InvalidParameter ("__init__: crypto parameter version number "
                                    "must be an integer, not %s"
                                    % type (paramversion))
        if paramversion < 0:
            raise InvalidParameter ("__init__: crypto parameter version number "
                                    "must be a nonnegative integer, not %d"
                                    % paramversion)
        # salt
        if nacl is not None:
            if isinstance (nacl, bytes) is False:
                raise InvalidParameter ("__init__: salt given, but of type %s "
                                        "instead of bytes" % type (nacl))
            # salt length would depend on the actual encryption so it can’t be
            # validated at this point
        self.fixed        = [ ]
        self.version      = version
        self.paramenc     = ENCRYPTION_PARAMETERS.get (paramversion) ["enc"]

        super().__init__ (password, key, paramversion, nacl, counter=counter,
                          nextpfx=lambda: self.fixed.append (os.urandom(8)),
                          strict_ivs=strict_ivs)


    def iv_make (self):
        return struct.pack(FMT_I2N_IV, self.fixed [-1], self.cnt)


    def next (self, filename=None, counter=None):
        if filename is None:
            if self.lastinfo is None:
                raise InvalidParameter ("next: filename is mandatory for "
                                        "first object")
            filename, _dummy = self.lastinfo
        else:
            if isinstance (filename, str) is False:
                raise InvalidParameter ("next: filename must be a string, no %s"
                                        % type (filename))
        if counter is not None:
            if isinstance (counter, int) is False:
                raise InvalidParameter ("next: the supplied counter is of "
                                        "invalid type %s; please pass an "
                                        "integer instead" % type (counter))
            self.set_object_counter (counter)
        self.iv = self.iv_make ()
        if self.paramenc == "aes-gcm":
            self.enc = Cipher \
                            ( algorithms.AES (self.key)
                            , modes.GCM (self.iv)
                            , backend = default_backend ()) \
                            .encryptor ()
        elif self.paramenc == "passthrough":
            self.enc = PassthroughCipher ()
        else:
            raise InvalidParameter ("next: parameter version %d not known"
                                    % self.paramversion)
        hdrdum = hdr_make_dummy (filename)
        self.lastinfo = (filename, hdrdum)
        super().next (self.password, self.paramversion, self.nacl, self.iv)

        self.set_object_counter (self.cnt + 1)
        return hdrdum


    def done (self, cmpdata):
        if isinstance (cmpdata, bytes) is False:
            raise InvalidParameter ("done: comparison input expected as bytes, "
                                    "not %s" % type (cmpdata))
        if self.lastinfo is None:
            raise RuntimeError ("done: encryption context not initialized")
        filename, hdrdum = self.lastinfo
        if cmpdata != hdrdum:
            raise RuntimeError ("done: bad sync of header for object %d: "
                                "preliminary data does not match; this likely "
                                "indicates a wrongly repositioned stream"
                                % self.cnt)
        data = self.enc.finalize ()
        self.stats ["out"] += len (data)
        self.ctsize += len (data)
        ok, hdr = hdr_from_params (self.version, self.paramversion, self.nacl,
                                   self.iv, self.ctsize, self.enc.tag)
        if ok is False:
            raise InternalError ("error constructing header: %r" % hdr)
        return data, hdr, self.fixed


    def process (self, buf):
        if isinstance (buf, bytes) is False:
            raise InvalidParameter ("process: expected byte buffer, not %s"
                                    % type (buf))
        bsize = len (buf)
        newptsize = self.ptsize + bsize
        diff = newptsize - PDTCRYPT_MAX_OBJ_SIZE
        if diff > 0:
            bsize -= diff
            newptsize = PDTCRYPT_MAX_OBJ_SIZE
        self.ptsize = newptsize
        data = super().process (buf [:bsize])
        self.ctsize += len (data)
        return bsize, data


class Decrypt (Crypto):

    tag        = None   # GCM tag, part of header
    last_iv    = None   # check consecutive ivs in strict mode

    def __init__ (self, password=None, key=None, counter=None, fixedparts=None,
                  strict_ivs=False):
        if         password is     None and key is     None \
                or password is not None and key is not None :
            raise InvalidParameter ("__init__: need either key or password")

        if key is not None:
            if isinstance (key, bytes) is False:
                raise InvalidParameter ("__init__: key must be provided as "
                                        "bytes, not %s" % type (key))
        else: # password, no key
            if isinstance (password, str) is False:
                raise InvalidParameter ("__init__: password must be a string, not %s"
                                        % type (password))
            if len (password) == 0:
                raise InvalidParameter ("__init__: supplied empty password but not "
                                        "permitted for PDT encrypted files")
        # fixed parts
        if fixedparts is not None:
            if isinstance (fixedparts, list) is False:
                raise InvalidParameter ("__init__: IV fixed parts must be "
                                        "supplied as list, not %s"
                                        % type (fixedparts))
            self.fixed = fixedparts
            self.fixed.sort ()

        super().__init__ (password, key, counter=counter, strict_ivs=strict_ivs)


    def valid_fixed_part (self, iv):
        # check if fixed part is known
        fixed, _cnt = struct.unpack (FMT_I2N_IV, iv)
        i = bisect.bisect_left (self.fixed, fixed)
        return i != len (self.fixed) and self.fixed [i] == fixed


    def check_consecutive_iv (self, iv):
        fixed, cnt = struct.unpack (FMT_I2N_IV, iv)
        if self.strict_ivs is True \
                and self.last_iv is not None \
                and self.last_iv [0] == fixed \
                and self.last_iv [1] != cnt - 1:
            raise NonConsecutiveIV ("iv %s counter not successor of "
                                    "last object (expected %d, found %d)"
                                    % (iv_fmt (self.last_iv [1]), cnt))
        self.last_iv = (iv, cnt)


    def next (self, hdr):
        if isinstance (hdr, bytes) is True:
            hdr = hdr_read (hdr)
        elif isinstance (hdr, dict) is False:
            # this won’t catch malformed specs though
            raise InvalidParameter ("next: wrong type of parameter hdr: "
                                    "expected bytes or spec, got %s"
                                    % type (hdr))
        try:
            paramversion = hdr ["paramversion"]
            nacl         = hdr ["nacl"]
            iv           = hdr ["iv"]
            tag          = hdr ["tag"]
        except KeyError:
            raise InvalidHeader ("next: not a header %r" % hdr)

        super().next (self.password, paramversion, nacl, iv)
        if self.fixed is not None and self.valid_fixed_part (iv) is False:
            raise InvalidIVFixedPart ("iv %s has invalid fixed part"
                                      % iv_fmt (iv))
        self.check_consecutive_iv (iv)

        self.tag = tag
        defs = ENCRYPTION_PARAMETERS.get (paramversion, None)
        if defs is None:
            raise FormatError ("header contains unknown parameter version %d; "
                               "maybe the file was created by a more recent "
                               "version of Deltatar" % paramversion)
        enc = defs ["enc"]
        if enc == "aes-gcm":
            self.enc = Cipher \
                            ( algorithms.AES (self.key)
                            , modes.GCM (iv, tag=self.tag)
                            , backend = default_backend ()) \
                            . decryptor ()
        elif enc == "passthrough":
            self.enc = PassthroughCipher ()
        else:
            raise InternalError ("encryption parameter set %d refers to unknown "
                                 "mode %r" % (paramversion, enc))
        self.set_object_counter (self.cnt + 1)


    def done (self, tag=None):
        data = b""
        try:
            if tag is None:
                data = self.enc.finalize ()
            else:
                if isinstance (tag, bytes) is False:
                    raise InvalidParameter ("done: wrong type of parameter "
                                            "tag: expected bytes, got %s"
                                            % type (tag))
                data = self.enc.finalize_with_tag (self.tag)
        except cryptography.exceptions.InvalidTag:
            raise InvalidGCMTag ("done: tag mismatch of object %d: %s "
                                  "rejected by finalize ()"
                                  % (self.cnt, binascii.hexlify (self.tag)))
        self.ctsize += len (data)
        self.stats ["out"] += len (data)
        return data


    def process (self, buf):
        if isinstance (buf, bytes) is False:
            raise InvalidParameter ("process: expected byte buffer, not %s"
                                    % type (buf))
        self.ctsize += len (buf)
        data = super().process (buf)
        self.ptsize += len (data)
        return data


###############################################################################
## testing helpers
###############################################################################

def _patch_global (glob, vow, n=None):
    """
    Adapt upper file counter bound for testing IV logic. Completely unsafe.
    """
    assert vow == "I am fully aware that this will void my warranty."
    r = globals () [glob]
    if n is None:
        n = globals () [glob + "_DEFAULT"]
    globals () [glob] = n
    return r

_testing_set_AES_GCM_IV_CNT_MAX = \
        partial (_patch_global, "AES_GCM_IV_CNT_MAX")

_testing_set_PDTCRYPT_MAX_OBJ_SIZE = \
        partial (_patch_global, "PDTCRYPT_MAX_OBJ_SIZE")

###############################################################################
## freestanding invocation
###############################################################################

PDTCRYPT_SUB_PROCESS = 0
PDTCRYPT_SUB_SCRYPT  = 1

PDTCRYPT_SUB = \
        { "process" : PDTCRYPT_SUB_PROCESS
        , "scrypt"  : PDTCRYPT_SUB_SCRYPT }

PDTCRYPT_DECRYPT   = 1 << 0 # decrypt archive with password
PDTCRYPT_SPLIT     = 1 << 1 # split archive into individual objects
PDTCRYPT_HASH      = 1 << 2 # output scrypt hash for file and given password

PDTCRYPT_SPLITNAME = "pdtcrypt-object-%d.bin"

PDTCRYPT_VERBOSE   = False
PDTCRYPT_STRICTIVS = False
PDTCRYPT_OVERWRITE = False
PDTCRYPT_BLOCKSIZE = 1 << 12
PDTCRYPT_SINK      = 0
PDTCRYPT_SOURCE    = 1
SELF               = None


class PDTDecryptionError (Exception):
    """Decryption failed."""

class PDTSplitError (Exception):
    """Decryption failed."""


def noise (*a, **b):
    print (file=sys.stderr, *a, **b)


class PassthroughDecryptor (object):

    curhdr = None # write current header on first data write

    def __init__ (self):
        if PDTCRYPT_VERBOSE is True:
            noise ("PDT: no encryption; data passthrough")

    def next (self, hdr):
        ok, curhdr = hdr_make (hdr)
        if ok is False:
            raise PDTDecryptionError ("bad header %r" % hdr)
        self.curhdr = curhdr

    def done (self):
        if self.curhdr is not None:
            return self.curhdr
        return b""

    def process (self, d):
        if self.curhdr is not None:
            d = self.curhdr + d
            self.curhdr = None
        return d


def depdtcrypt (mode, pw, ins, outs):
    """
    Remove PDTCRYPT layer from obj encrypted with pw. Used on a Deltatar
    backup this will yield a (possibly Gzip compressed) tarball.
    """
    ctleft     = -1              # length of ciphertext to consume
    ctcurrent  = 0               # total ciphertext of current object
    total_obj  = 0               # total number of objects read
    total_pt   = 0               # total plaintext bytes
    total_ct   = 0               # total ciphertext bytes
    total_read = 0               # total bytes read
    outfile    = None            # Python file object for output

    if mode & PDTCRYPT_DECRYPT:  # decryptor
        decr = Decrypt (password=pw, strict_ivs=PDTCRYPT_STRICTIVS)
    else:
        decr = PassthroughDecryptor ()

    def nextout (_):
        """Dummy for non-split mode: output file does not vary."""
        return outs

    if mode & PDTCRYPT_SPLIT:
        def nextout (outfile):
            """
            We were passed an fd as outs for accessing the destination
            directory where extracted archive components are supposed
            to end up in.
            """

            if outfile is None:
                if PDTCRYPT_VERBOSE is True:
                    noise ("PDT: no output file to close at this point")
            else:
                if PDTCRYPT_VERBOSE is True:
                    noise ("PDT: release output file %r" % outfile)
                    # cleanup happens automatically by the GC; the next
                    # line will error out on account of an invalid fd
                    #outfile.close ()

            assert total_obj > 0
            fname = PDTCRYPT_SPLITNAME % total_obj
            try:
                oflags = os.O_CREAT | os.O_WRONLY
                if PDTCRYPT_OVERWRITE is True:
                    oflags |= os.O_TRUNC
                else:
                    oflags |= os.O_EXCL
                outfd = os.open (fname, oflags, 0o600, dir_fd=outs)
                if PDTCRYPT_VERBOSE is True:
                    noise ("PDT: new output file %s → %d" % (fname, outfd))
            except FileExistsError as exn:
                noise ("PDT: refusing to overwrite existing file %s" % fname)
                noise ("")
                raise PDTSplitError ("destination file %s already exists"
                                     % fname)

            return os.fdopen (outfd, "wb", closefd=True)


    def tell (s):
        """ESPIPE is normal on stdio stream."""
        try:
            return s.tell ()
        except OSError as exn:
            if exn.errno == 29:
                return -1

    def out (pt, outfile):
        npt = len (pt)
        nonlocal total_pt
        total_pt += npt
        if PDTCRYPT_VERBOSE is True:
            noise ("PDT:\t· decrypt plaintext %d B" % (npt))
        try:
            nn = outfile.write (pt)
        except OSError as exn: # probably ENOSPC
            raise DecryptionError ("error (%s)" % exn)
        if nn != npt:
            raise DecryptionError ("write aborted after %d of %d B" % (nn, npt))

    while True:
        if ctleft <= 0:
            # current object completed; in a valid archive this marks either
            # the start of a new header or the end of the input
            if ctleft == 0: # current object requires finalization
                if PDTCRYPT_VERBOSE is True:
                    noise ("PDT: %d finalize" % tell (ins))
                try:
                    pt = decr.done ()
                except InvalidGCMTag as exn:
                    raise DecryptionError ("error finalizing object %d (%d B): "
                                           "%r" % (total_obj, len (pt), exn)) \
                          from exn
                out (pt, outfile)
                if PDTCRYPT_VERBOSE is True:
                    noise ("PDT:\t· object validated")

            if PDTCRYPT_VERBOSE is True:
                noise ("PDT: %d hdr" % tell (ins))
            try:
                hdr = hdr_read_stream (ins)
                total_read += PDTCRYPT_HDR_SIZE
            except EndOfFile as exn:
                total_read += exn.remainder
                if total_ct + total_obj * PDTCRYPT_HDR_SIZE != total_read:
                    raise PDTDecryptionError ("ciphertext processed (%d B) plus "
                                              "overhead (%d × %d B) does not match "
                                              "the number of bytes read (%d )"
                                              % (total_ct, total_obj, PDTCRYPT_HDR_SIZE,
                                                 total_read))
                # the single good exit
                return total_read, total_obj, total_ct, total_pt
            except InvalidHeader as exn:
                raise PDTDecryptionError ("invalid header at position %d in %r "
                                          "(%s)" % (tell (ins), exn, ins))
            if PDTCRYPT_VERBOSE is True:
                pretty = hdr_fmt_pretty (hdr)
                noise (reduce (lambda a, e: (a + "\n" if a else "") + "PDT:\t· " + e,
                               pretty.splitlines (), ""))
            ctcurrent = ctleft = hdr ["ctsize"]

            decr.next (hdr)

            total_obj += 1 # used in file counter with split mode

            # finalization complete or skipped in case of first object in
            # stream; create a new output file if necessary
            outfile = nextout (outfile)

            if PDTCRYPT_VERBOSE is True:
                noise ("PDT: %d decrypt obj no. %d,  %d B"
                       % (tell (ins), total_obj, ctleft))

        # always allocate a new buffer since python-cryptography doesn’t allow
        # passing a bytearray :/
        nexpect = min (ctleft, PDTCRYPT_BLOCKSIZE)
        if PDTCRYPT_VERBOSE is True:
            noise ("PDT:\t· [%d] %d%% done, read block (%d B of %d B remaining)"
                   % (tell (ins),
                      100 - ctleft * 100 / (ctcurrent > 0 and ctcurrent or 1),
                      nexpect, ctleft))
        ct      = ins.read (nexpect)
        nct     = len (ct)
        if nct < nexpect:
            off = tell (ins)
            raise EndOfFile (nct,
                             "hit EOF after %d of %d B in block [%d:%d); "
                             "%d B ciphertext remaining for object no %d"
                             % (nct, nexpect, off, off + nexpect, ctleft,
                                total_obj))
        ctleft     -= nct
        total_ct   += nct
        total_read += nct

        if PDTCRYPT_VERBOSE is True:
            noise ("PDT:\t· decrypt ciphertext %d B" % (nct))
        pt = decr.process (ct)
        out (pt, outfile)


def deptdcrypt_mk_stream (kind, path):
    """Create stream from file or stdio descriptor."""
    if kind == PDTCRYPT_SINK:
        if path == "-":
            if PDTCRYPT_VERBOSE is True: noise ("PDT: sink: stdout")
            return sys.stdout.buffer
        else:
            if PDTCRYPT_VERBOSE is True: noise ("PDT: sink: file %s" % path)
            return io.FileIO (path, "w")
    if kind == PDTCRYPT_SOURCE:
        if path == "-":
            if PDTCRYPT_VERBOSE is True: noise ("PDT: source: stdin")
            return sys.stdin.buffer
        else:
            if PDTCRYPT_VERBOSE is True: noise ("PDT: source: file %s" % path)
            return io.FileIO (path, "r")

    raise ValueError ("bogus stream “%s” / %s" % (kind, path))


def depdtcrypt_file (pw, spath, dpath):
    """
    Remove PDTCRYPT layer from file at ``spath`` using password ``pw``, writing
    the decrypted result to dpath.
    """
    if PDTCRYPT_VERBOSE is True:
        noise ("PDT: decrypt %s → %s" % (spath, dpath), file=sys.stderr)
    with deptdcrypt_mk_stream (PDTCRYPT_SOURCE, spath) as ins:
        with deptdcrypt_mk_stream (PDTCRYPT_SINK, dpath) as outs:
            return depdtcrypt (pw, ins, outs)


def mode_depdtcrypt (mode, pw, ins, outs):
    try:
        total_read, total_obj, total_ct, total_pt = \
            depdtcrypt (mode, pw, ins, outs)
    except DecryptionError as exn:
        noise ("PDT: Decryption failed:")
        noise ("PDT:")
        noise ("PDT:    “%s”" % exn)
        noise ("PDT:")
        noise ("PDT: Did you specify the correct password?")
        noise ("")
        return 1
    except PDTSplitError as exn:
        noise ("PDT: Split operation failed:")
        noise ("PDT:")
        noise ("PDT:    “%s”" % exn)
        noise ("PDT:")
        noise ("PDT: Hint: target directory should to be empty.")
        noise ("")
        return 1

    if PDTCRYPT_VERBOSE is True:
        noise ("PDT: decryption successful"                 )
        noise ("PDT:   %.10d bytes read"        % total_read)
        noise ("PDT:   %.10d objects decrypted" % total_obj )
        noise ("PDT:   %.10d bytes ciphertext"  % total_ct  )
        noise ("PDT:   %.10d bytes plaintext"   % total_pt  )
        noise (""                                           )

    return 0


def mode_scrypt (pw, ins):
    hsh, nacl, params = scrypt_hashsource (pw, ins)

    import json
    out = json.dumps ({ "salt"          : str (binascii.hexlify (nacl))
                      , "hash"          : str (binascii.hexlify (hsh))
                      , "scrypt_params" : { "N"     : params ["N"]
                                          , "r"     : params ["r"]
                                          , "p"     : params ["p"]
                                          , "dkLen" : params ["dkLen"] } })
    print (out)


def usage (err=False):
    out = print
    if err is True:
        out = noise
    out ("usage: %s SUBCOMMAND { --help" % SELF)
    out ("                     | [ -v ] { PASSWORD } { { -i | --in }  { - | SOURCE } }")
    out ("                                           { { -o | --out } { - | DESTINATION } }")
    out ("                                           { -D | --no-decrypt } { -S | --split }")
    out ("")
    out ("\twhere")
    out ("\t\tSUBCOMMAND      main mode: { process | scrypt }")
    out ("\t\t                where:")
    out ("\t\t                   process: extract objects from PDT archive")
    out ("\t\t                   scrypt:  calculate hash from password and first object")
    out ("\t\tPASSWORD        password to derive the encryption key from")
    out ("\t\t-s              enforce strict handling of initialization vectors")
    out ("\t\t-i SOURCE       file name to read from")
    out ("\t\t-o DESTINATION  file to write output to")
    out ("\t\t-v              print extra info")
    out ("\t\t-S              split into files at object boundaries; this")
    out ("\t\t                requires DESTINATION to refer to directory")
    out ("\t\t-D              PDT header and ciphertext passthrough")
    out ("")
    out ("\tinstead of filenames, “-” may used to specify stdin / stdout")
    out ("")
    sys.exit ((err is True) and 42 or 0)


def parse_argv (argv):
    global SELF
    mode      = PDTCRYPT_DECRYPT
    pw        = None
    insspec   = None
    outsspec  = None

    argvi = iter (argv)
    SELF  = os.path.basename (next (argvi))

    try:
        rawsubcmd  = next (argvi)
        subcommand = PDTCRYPT_SUB [rawsubcmd]
    except StopIteration:
        noise ("ERROR: subcommand required")
        noise ("")
        usage (err=True)
        raise Unreachable
    except KeyError:
        noise ("ERROR: invalid subcommand “%s” specified" % rawsubcmd)
        noise ("")
        usage (err=True)
        raise Unreachable

    def checked_pw (arg):
        nonlocal pw
        if pw is None:
            pw = arg
        else:
            noise ("ERROR: unqualified argument “%s” but password already "
                "given" % arg)
            noise ("")
            usage (err=True)
            raise Unreachable

    for arg in argvi:
        if arg in [ "-h", "--help" ]:
            usage ()
            raise Unreachable
        elif arg in [ "-v", "--verbose", "--wtf" ]:
            global PDTCRYPT_VERBOSE
            PDTCRYPT_VERBOSE = True
        elif arg in [ "-i", "--in", "--source" ]:
            insspec = next (argvi)
            if PDTCRYPT_VERBOSE is True: noise ("PDT: decrypt from %s" % insspec)
        elif arg in [ "-o", "--out", "--dest", "--sink" ]:
            outsspec = next (argvi)
            if PDTCRYPT_VERBOSE is True: noise ("PDT: decrypt to %s" % outsspec)
        else:
            if subcommand == PDTCRYPT_SUB_PROCESS:
                if arg in [ "-s", "--strict-ivs" ]:
                    global PDTCRYPT_STRICTIVS
                    PDTCRYPT_STRICTIVS = True
                elif arg in [ "-f", "--force" ]:
                    global PDTCRYPT_OVERWRITE
                    PDTCRYPT_OVERWRITE = True
                    if PDTCRYPT_VERBOSE is True: noise ("PDT: overwrite existing files")
                elif arg in [ "-S", "--split" ]:
                    mode |= PDTCRYPT_SPLIT
                    if PDTCRYPT_VERBOSE is True: noise ("PDT: split files")
                elif arg in [ "-D", "--no-decrypt" ]:
                    mode &= ~PDTCRYPT_DECRYPT
                    if PDTCRYPT_VERBOSE is True: noise ("PDT: not decrypting")
                else:
                    checked_pw (arg)
            elif subcommand == PDTCRYPT_SUB_SCRYPT:
                checked_pw (arg)
            else:
                raise Unreachable

    if pw is None:
        if subcommand == PDTCRYPT_SUB_SCRYPT:
            noise ("ERROR: scrypt hash mode requested but no password given")
            noise ("")
            usage (err=True)
            raise Unreachable
        elif mode & PDTCRYPT_DECRYPT:
            noise ("ERROR: encryption requested but no password given")
            noise ("")
            usage (err=True)
            raise Unreachable

    # default to stdout
    ins = deptdcrypt_mk_stream (PDTCRYPT_SOURCE, insspec  or "-")

    if subcommand == PDTCRYPT_SUB_SCRYPT:
        return True, partial (mode_scrypt, pw.encode (), ins)

    if mode & PDTCRYPT_SPLIT: # destination must be directory
        if outsspec is None or outsspec == "-":
            noise ("ERROR: split mode is incompatible with stdout sink")
            noise ("")
            usage (err=True)
            raise Unreachable

        try:
            try:
                os.makedirs (outsspec, 0o700)
            except FileExistsError:
                # if it’s a directory with appropriate perms, everything is
                # good; otherwise, below invocation of open(2) will fail
                pass
            outs = os.open (outsspec, os.O_DIRECTORY, 0o600)
        except FileNotFoundError as exn:
            noise ("ERROR: cannot create target directory “%s”" % outsspec)
            noise ("")
            usage (err=True)
            raise Unreachable
        except NotADirectoryError as exn:
            noise ("ERROR: target path “%s” is not a directory" % outsspec)
            noise ("")
            usage (err=True)
            raise Unreachable

    else:
        outs = deptdcrypt_mk_stream (PDTCRYPT_SINK, outsspec or "-")

    return True, partial (mode_depdtcrypt, mode, pw, ins, outs)


def main (argv):
    ok, runner = parse_argv (argv)

    if ok is True: return runner ()

    return 1


if __name__ == "__main__":
    sys.exit (main (sys.argv))

